#!/usr/bin/env node
"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const chokidar = __importStar(require("chokidar"));
const logSymbols = __importStar(require("log-symbols"));
const meow_1 = __importDefault(require("meow"));
const path = __importStar(require("path"));
const errors_1 = require("./errors");
const format = __importStar(require("./format"));
const parser_1 = require("./parser");
const validation_1 = require("./validation");
const cli = meow_1.default(`
  Usage
    $ postguard ./path/to/source/*.ts

  Options
    --help        Print this help
    -w, --watch   Watch files and re-evaluate on change
`, {
    flags: {
        help: { type: "boolean" },
        w: { type: "boolean" },
        watch: { type: "boolean" }
    }
});
if (cli.input.length === 0) {
    cli.showHelp();
    process.exit(0);
}
const watchMode = Boolean(cli.flags.watch || cli.flags.w);
function checkSchemasForDuplicates(allTableSchemas) {
    for (const schema of allTableSchemas) {
        const schemasMatchingThatName = allTableSchemas.filter(someSchema => someSchema.tableName === schema.tableName);
        if (schemasMatchingThatName.length > 1) {
            throw new Error(`Table "${schema.tableName}" has been defined more than once:\n` +
                schemasMatchingThatName
                    .map(duplicate => {
                    const lineRef = duplicate.loc ? `:${duplicate.loc.start.line}` : ``;
                    return `  - ${duplicate.sourceFile.filePath}${lineRef}`;
                })
                    .join("\n"));
        }
    }
}
function run(sourceFilePaths, moreSchemas = []) {
    let allQueries = [];
    let allTableSchemas = [...moreSchemas];
    sourceFilePaths = sourceFilePaths.map(filePath => path.relative(process.cwd(), filePath));
    try {
        for (const filePath of sourceFilePaths) {
            const { queries, tableSchemas } = parser_1.parseSourceFile(parser_1.loadSourceFile(filePath));
            // TODO: Check that no table is defined twice
            allQueries = [...allQueries, ...queries];
            allTableSchemas = [...allTableSchemas, ...tableSchemas];
        }
        checkSchemasForDuplicates(allTableSchemas);
        for (const query of allQueries) {
            validation_1.validateQuery(query, allTableSchemas);
        }
        console.log(format.success(`${logSymbols.success} Validated ${allQueries.length} queries against ${allTableSchemas.length} table schemas. All fine!`));
    }
    catch (error) {
        if (errors_1.isAugmentedError(error)) {
            const [firstLine, ...lines] = String(error.message).split("\n");
            const linesIndented = lines.map(line => (line.match(/^>?\s*\d*\s*\|/) ? line : `  ${line}`));
            console.error(`${logSymbols.error} ${[firstLine, ...linesIndented].join("\n")}`);
        }
        else {
            console.error(error.stack);
        }
        if (!watchMode) {
            process.exit(1);
        }
    }
    return {
        queries: allQueries,
        schemas: allTableSchemas
    };
}
let { schemas } = run(cli.input);
if (watchMode) {
    console.log("\nWatching file changes... Press CTRL + C to cancel.");
    chokidar.watch(cli.input, { ignoreInitial: true }).on("all", (event, filePath) => {
        console.log(format.gray(`\nRe-running after file changed: ${filePath}\n`));
        const schemasInOtherFiles = schemas.filter(schema => schema.sourceFile.filePath !== filePath);
        const lastRunResult = run([filePath], schemasInOtherFiles);
        schemas = lastRunResult.schemas;
    });
}
