"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const code_frame_1 = require("@babel/code-frame");
const format = __importStar(require("./format"));
function fail(message) {
    throw new Error(message);
}
exports.fail = fail;
function formatSourceLink(filePath, location) {
    if (location) {
        const locationString = location.start.column
            ? `${filePath}:${location.start.line}:${location.start.column + 1}`
            : `${filePath}:${location.start.line}`;
        return format.sourceReference(locationString);
    }
    else {
        return format.sourceReference(filePath);
    }
}
function getLineColumnOffset(text, startIndex) {
    const preceedingText = text.substring(0, startIndex);
    const lineOffset = preceedingText.replace(/[^\n]/g, "").length;
    const columnOffset = startIndex - preceedingText.lastIndexOf("\n");
    return {
        columnOffset,
        lineOffset
    };
}
function getOverallSourceLocation(locations) {
    const first = locations[0];
    const last = locations[locations.length - 1];
    return {
        start: {
            column: first.start.column,
            line: first.start.line
        },
        end: {
            column: last.end ? last.end.column : last.start.column,
            line: last.end ? last.end.line : last.start.line
        }
    };
}
function mapToSourceLocation(query, stringIndex) {
    const matchingSpan = query.sourceMap.find(span => span.queryStartIndex <= stringIndex && span.queryEndIndex > stringIndex);
    if (!matchingSpan) {
        return getOverallSourceLocation(query.sourceMap.map(span => span.sourceLocation));
    }
    const indexInSpan = stringIndex - matchingSpan.queryStartIndex;
    const textInSpan = query.query.substring(matchingSpan.queryStartIndex, matchingSpan.queryEndIndex);
    const offsetsInSpan = getLineColumnOffset(textInSpan, indexInSpan);
    const columnOffsetInSpan = matchingSpan.isTemplateExpression ? 1 : offsetsInSpan.columnOffset;
    return {
        start: {
            column: offsetsInSpan.lineOffset > 0
                ? columnOffsetInSpan
                : matchingSpan.sourceLocation.start.column + columnOffsetInSpan,
            line: matchingSpan.sourceLocation.start.line + offsetsInSpan.lineOffset
        }
    };
}
function isAugmentedError(error) {
    return error.name === "QuerySyntaxError" || error.name === "ValidationError";
}
exports.isAugmentedError = isAugmentedError;
function augmentFileValidationError(error, query) {
    const location = "location" in error && error.location
        ? error.location
        : getOverallSourceLocation(query.sourceMap.map(span => span.sourceLocation));
    const formattedQuery = code_frame_1.codeFrameColumns(query.sourceFile.fileContent, location);
    error.name = "ValidationError";
    error.message =
        format.error(`Query validation failed in ${formatSourceLink(query.sourceFile.filePath, location)}:`) +
            `\n\n` +
            format.error(`${error.message}`) +
            `\n\n` +
            formattedQuery;
    return error;
}
exports.augmentFileValidationError = augmentFileValidationError;
function augmentQuerySyntaxError(error, syntaxError, query) {
    // tslint:disable-next-line prefer-object-spread
    return Object.assign(error, {
        name: "QuerySyntaxError",
        location: mapToSourceLocation(query, syntaxError.cursorPosition - 1),
        sourceFile: query.sourceFile
    });
}
exports.augmentQuerySyntaxError = augmentQuerySyntaxError;
function augmentValidationError(error, path, query) {
    const location = path.type in path.node && "location" in path.node[path.type]
        ? mapToSourceLocation(query, path.node[path.type].location)
        : getOverallSourceLocation(query.sourceMap.map(span => span.sourceLocation));
    // tslint:disable-next-line prefer-object-spread
    return Object.assign(error, {
        name: "ValidationError",
        path,
        location,
        sourceFile: query.sourceFile
    });
}
exports.augmentValidationError = augmentValidationError;
function augmentCodeError(error, path, query) {
    // tslint:disable-next-line prefer-object-spread
    return Object.assign(error, {
        name: "ValidationError",
        location: path.node.loc,
        sourceFile: query.sourceFile
    });
}
exports.augmentCodeError = augmentCodeError;
