"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const QueryParser = __importStar(require("pg-query-parser"));
const errors_1 = require("../errors");
const query_traversal_1 = require("./query-traversal");
const resolve_references_1 = require("./resolve-references");
const pg_node_types_1 = require("./pg-node-types");
const $any = Symbol("any");
exports.spreadTypeAny = Object.defineProperty({}, $any, {
    enumerable: false,
    value: true
});
const isPlaceholderSelect = (node) => pg_node_types_1.isSelectStmt(node) &&
    !node.SelectStmt.fromClause &&
    node.SelectStmt.targetList &&
    node.SelectStmt.targetList.length === 1 &&
    pg_node_types_1.isResTarget(node.SelectStmt.targetList[0]) &&
    pg_node_types_1.isParamRef(node.SelectStmt.targetList[0].ResTarget.val);
const isSubquery = (node) => pg_node_types_1.getNodeType(node).endsWith("Stmt") && !isPlaceholderSelect(node);
const isReturningIntoParentQuery = (path) => ["larg", "rarg", "selectStmt"].indexOf(path.parentPropKey) > -1;
function getSubqueries(statementPath) {
    const subqueries = [];
    query_traversal_1.traverseSubTree(statementPath, (path, $cancelRecursion) => {
        if (path.node === statementPath.node)
            return;
        if (isSubquery(path.node)) {
            subqueries.push(path);
            return $cancelRecursion;
        }
    });
    return subqueries;
}
function getTableReferences(statementPath, includeSubQueries) {
    const referencedTables = [];
    query_traversal_1.traverseSubTree(statementPath, (path, $cancelRecursion) => {
        const { node } = path;
        if (pg_node_types_1.isRelationRef(node)) {
            referencedTables.push(path);
        }
        else if (!includeSubQueries) {
            if (node !== statementPath.node && isSubquery(node))
                return $cancelRecursion;
        }
    });
    return referencedTables;
}
function getReferencedColumns(statement, spreadTypes) {
    const referencedColumns = [];
    query_traversal_1.traverseSubTree(statement, (path, $cancelRecursion) => {
        const spreadType = pg_node_types_1.isParamRef(path.node) ? spreadTypes[path.node.ParamRef.number] : null;
        if (path.node === statement.node)
            return;
        if (isSubquery(path.node))
            return $cancelRecursion;
        if (pg_node_types_1.isColumnRef(path.node)) {
            const relationRefs = getTableReferences(statement, false);
            const columnRef = resolve_references_1.resolveColumnReference(path, relationRefs);
            if (columnRef) {
                referencedColumns.push(columnRef);
            }
        }
        else if (pg_node_types_1.isResTarget(path.node) && path.node.ResTarget.name) {
            const relationRefs = getTableReferences(statement, false);
            const columnRef = resolve_references_1.resolveResTarget(path, relationRefs);
            if (columnRef) {
                referencedColumns.push(columnRef);
            }
        }
        else if (spreadType) {
            const relationRefs = getTableReferences(statement, false);
            const tableRefsInScope = relationRefs.map(ref => ({
                as: ref.node.RangeVar.alias ? ref.node.RangeVar.alias.Alias.aliasname : undefined,
                tableName: ref.node.RangeVar.relname,
                path
            }));
            for (const spreadArgKey of Object.keys(spreadType)) {
                referencedColumns.push({
                    tableRefsInScope: resolve_references_1.filterDuplicateTableRefs(tableRefsInScope),
                    columnName: spreadArgKey,
                    path
                });
            }
            if (spreadType === exports.spreadTypeAny || !spreadType) {
                referencedColumns.push({
                    any: true,
                    tableRefsInScope: resolve_references_1.filterDuplicateTableRefs(tableRefsInScope),
                    columnName: "",
                    path
                });
            }
        }
    });
    return referencedColumns;
}
function getReturningColumns(statement) {
    const body = statement.node[statement.type];
    const relationRefs = getTableReferences(statement, false);
    const { returningList = [], targetList = [] } = body;
    const resTargets = [...returningList, ...targetList].filter(node => pg_node_types_1.getNodeType(node) === "ResTarget");
    return resTargets
        .map(resTarget => {
        const propKey = returningList.indexOf(resTarget) > -1 ? "returningList" : "targetList";
        const resTargetPath = query_traversal_1.createQueryNodeSubpath(statement, resTarget, propKey);
        return resolve_references_1.resolveResTarget(resTargetPath, relationRefs);
    })
        .filter((ref) => ref !== null);
}
function instantiateQuery(path, context) {
    const referencedColumns = getReferencedColumns(path, context.expressionSpreadTypes);
    const referencedTables = getTableReferences(path, true).map(tableRef => ({
        tableName: tableRef.node.RangeVar.relname,
        path: tableRef
    }));
    const returnedColumns = getReturningColumns(path);
    const subqueries = getSubqueries(path).map(subqueryPath => instantiateQuery(subqueryPath, context));
    const type = pg_node_types_1.getNodeType(path.node)
        .replace(/Stmt$/, "")
        .toUpperCase();
    const parent = query_traversal_1.getQueryPathParent(path);
    const exposedAsTable = parent && pg_node_types_1.isCommonTableExpr(parent.node) ? parent.node.CommonTableExpr.ctename : undefined;
    return {
        type,
        path,
        exposedAsTable,
        referencedColumns,
        referencedTables,
        returnedColumns,
        returnsIntoParentQuery: isReturningIntoParentQuery(path),
        query: context.query,
        sourceFile: context.sourceFile,
        sourceMap: context.sourceMap,
        subqueries
    };
}
function parsePostgresQuery(queryString, sourceFile, sourceMap, spreadTypes) {
    const context = {
        expressionSpreadTypes: spreadTypes,
        query: queryString,
        sourceFile,
        sourceMap
    };
    const result = QueryParser.parse(queryString);
    if (result.error) {
        const fakePath = query_traversal_1.createQueryNodePath({ SelectStmt: { op: 0 } }, [], "");
        const query = instantiateQuery(fakePath, context);
        const error = new Error(`Syntax error in SQL query.\nSubstituted query: ${queryString.trim()}`);
        throw errors_1.augmentFileValidationError(errors_1.augmentQuerySyntaxError(error, result.error, query), query);
    }
    const parsedQuery = result.query[0];
    const queryPath = query_traversal_1.createQueryNodePath(parsedQuery, [], "");
    return instantiateQuery(queryPath, context);
}
exports.parsePostgresQuery = parsePostgresQuery;
