"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const query_traversal_1 = require("./query-traversal");
const pg_node_types_1 = require("./pg-node-types");
function filterDuplicateTableRefs(tableRefs) {
    return tableRefs.reduce((filtered, ref) => filtered.find(someRef => JSON.stringify(someRef) === JSON.stringify(ref))
        ? filtered
        : [...filtered, ref], []);
}
exports.filterDuplicateTableRefs = filterDuplicateTableRefs;
function resolveTableName(tableIdentifier, tableRefs) {
    const matchingTableRef = tableRefs.find(tableRef => {
        const refNode = tableRef.node;
        return refNode.RangeVar.alias
            ? refNode.RangeVar.alias.Alias.aliasname === tableIdentifier
            : refNode.RangeVar.relname === tableIdentifier;
    });
    if (matchingTableRef) {
        return matchingTableRef.node.RangeVar.relname;
    }
    else {
        throw new Error(`No matching table reference found for "${tableIdentifier}".`);
    }
}
exports.resolveTableName = resolveTableName;
function resolveColumnReference(path, relationRefs) {
    const { fields } = path.node.ColumnRef;
    if (fields.length === 1) {
        const [columnNode] = fields;
        if (pg_node_types_1.isPgString(columnNode)) {
            const tableRefsInScope = relationRefs.map(ref => ({
                as: ref.node.RangeVar.alias ? ref.node.RangeVar.alias.Alias.aliasname : undefined,
                tableName: ref.node.RangeVar.relname,
                path
            }));
            return {
                tableRefsInScope: filterDuplicateTableRefs(tableRefsInScope),
                columnName: columnNode.String.str,
                path
            };
        }
        else if (pg_node_types_1.isStar(columnNode)) {
            if (relationRefs.length !== 1) {
                throw new Error(`Can only have unqualified * selector if only one table is referenced in the (sub-)query.\n` +
                    `Tables in scope: ${relationRefs.length === 0
                        ? "(none)"
                        : relationRefs.map(ref => ref.node.RangeVar.relname)}`);
            }
            const tableRef = relationRefs[0];
            return {
                tableName: tableRef.node.RangeVar.alias
                    ? tableRef.node.RangeVar.alias.Alias.aliasname
                    : tableRef.node.RangeVar.relname,
                columnName: "*",
                path
            };
        }
    }
    else if (fields.length === 2) {
        const [tableNode, columnNode] = fields;
        if (!pg_node_types_1.isPgString(tableNode)) {
            throw new Error(`Expected first identifier in column reference to be a string. Got ${pg_node_types_1.getNodeType(tableNode)}`);
        }
        if (pg_node_types_1.isPgString(columnNode)) {
            return {
                tableName: resolveTableName(tableNode.String.str, relationRefs),
                columnName: columnNode.String.str,
                path
            };
        }
        else if (pg_node_types_1.isStar(columnNode)) {
            return {
                tableName: resolveTableName(tableNode.String.str, relationRefs),
                columnName: "*",
                path
            };
        }
    }
    else {
        throw new Error(`Expected column reference to be of format <table>.<column> or <column>. Got: ${fields.join(".")}`);
    }
    return null;
}
exports.resolveColumnReference = resolveColumnReference;
function resolveResTarget(path, relationRefs) {
    const { name, val } = path.node.ResTarget;
    if (name) {
        const tableRefsInScope = relationRefs.map(ref => ({
            as: ref.node.RangeVar.alias ? ref.node.RangeVar.alias.Alias.aliasname : undefined,
            tableName: ref.node.RangeVar.relname,
            path
        }));
        return {
            tableRefsInScope: filterDuplicateTableRefs(tableRefsInScope),
            columnName: name,
            path
        };
    }
    else if (pg_node_types_1.isColumnRef(val)) {
        return resolveColumnReference(query_traversal_1.createQueryNodeSubpath(path, val, "val"), relationRefs);
    }
    else {
        return null;
    }
}
exports.resolveResTarget = resolveResTarget;
