"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const squid_1 = require("squid");
const errors_1 = require("./errors");
const resolver_1 = require("./resolver");
const utils_1 = require("./utils");
const isAllOfColumnReference = (ref) => ref.columnName === "*" || false;
const isUnresolvableColumnReference = (ref) => ref.any || false;
function assertIntactQualifiedColumnRef(columnRef, tables) {
    // Cannot validate column references that originate from non-inferrable spread expression
    if (isUnresolvableColumnReference(columnRef))
        return;
    const table = tables.find(someTable => someTable.tableName === columnRef.tableName);
    if (!table) {
        throw new Error(`No table named ${columnRef.tableName} has been defined.`);
    }
    if (table.columnNames.indexOf(columnRef.columnName) === -1) {
        throw new Error(`Table "${columnRef.tableName}" does not have a column named "${columnRef.columnName}".`);
    }
}
function assertIntactTableRef(tableRef, tables) {
    const tableSchema = tables.find(schema => schema.tableName === tableRef.tableName);
    if (!tableSchema) {
        throw new Error(`No table with name "${tableRef.tableName}" has been defined.`);
    }
    return tableSchema;
}
function assertNoBrokenColumnRefs(query, tables) {
    const referencedColumns = resolver_1.resolveColumnReferences(query, tables);
    for (const columnRef of referencedColumns) {
        try {
            const qualifiedColumnRef = "tableName" in columnRef ? columnRef : resolver_1.resolveUnqualifiedColumnRef(columnRef, tables);
            assertIntactQualifiedColumnRef(qualifiedColumnRef, tables);
        }
        catch (error) {
            throw errors_1.augmentValidationError(error, columnRef.path, query);
        }
    }
}
function assertNoBrokenTableRefs(query, tables) {
    for (const tableRef of query.referencedTables) {
        try {
            assertIntactTableRef(tableRef, tables);
        }
        catch (error) {
            throw errors_1.augmentValidationError(error, tableRef.path, query);
        }
    }
}
function assertCompleteInsertValues(query, tables) {
    // FIXME: Consider SELECT INTO queries as well
    if (query.type !== "INSERT")
        return;
    if (query.referencedColumns.some(isUnresolvableColumnReference))
        return;
    const schema = assertIntactTableRef(query.referencedTables[0], tables);
    const mandatoryColumns = schema.columnNames
        .map(columnName => (Object.assign({}, schema.columnDescriptors[columnName], { columnName })))
        .filter(descriptor => !descriptor.hasDefault);
    const queryColumns = [...query.referencedColumns, ...resolver_1.resolveColumnReferences(query, tables)];
    for (const mandatoryColumn of mandatoryColumns) {
        const columnReference = queryColumns.find(columnRef => columnRef.columnName === mandatoryColumn.columnName);
        if (!columnReference) {
            const error = new Error(`Column "${mandatoryColumn.columnName}" is missing from INSERT statement.`);
            throw errors_1.augmentValidationError(error, query.path, query);
        }
    }
}
function resolveColumnRefsToTableSchemaDescriptor(returnedColumns, tables) {
    let syntheticSchema = {};
    for (const columnRef of returnedColumns) {
        if (isAllOfColumnReference(columnRef)) {
            const referencedTableSchema = assertIntactTableRef(columnRef, tables);
            syntheticSchema = Object.assign({}, syntheticSchema, referencedTableSchema.columnDescriptors);
        }
        else {
            syntheticSchema[columnRef.columnName] = squid_1.Schema.Any;
        }
    }
    return syntheticSchema;
}
function validateQueryMatchesReturnType(resultSchema, expectedResult, path, query) {
    const expectedColumnNames = Object.keys(expectedResult);
    const actualColumnNames = Object.keys(resultSchema);
    const missingColumnNames = expectedColumnNames.filter(expectedColumnName => actualColumnNames.indexOf(expectedColumnName) === -1);
    if (missingColumnNames.length > 0) {
        const error = new Error(`Query's result does not match the expected result type.\n` +
            `  Missing columns in result rows: ${missingColumnNames
                .map(columnName => `"${columnName}"`)
                .join(", ")}\n` +
            `  Actual columns in result rows: ${actualColumnNames
                .map(columnName => `"${columnName}"`)
                .join(", ")}`);
        throw errors_1.augmentCodeError(error, path, query);
    }
    // TODO: Validate result column types
}
function resolveSubqueryToTableSchema(query, tables) {
    const schema = resolveColumnRefsToTableSchemaDescriptor(query.returnedColumns, tables);
    return {
        tableName: query.exposedAsTable,
        columnDescriptors: schema,
        columnNames: Object.keys(schema),
        loc: query.path.node.loc,
        sourceFile: query.sourceFile
    };
}
function validateSubquery(query, tables) {
    try {
        assertNoBrokenTableRefs(query, tables);
        assertNoBrokenColumnRefs(query, tables);
        assertCompleteInsertValues(query, tables);
    }
    catch (error) {
        throw errors_1.augmentFileValidationError(error, query);
    }
    for (const subquery of query.subqueries) {
        validateSubquery(subquery, tables);
    }
}
function validateQueryInvocation(invocation, tables) {
    if (!invocation.resultTypeAssertion ||
        invocation.query.returnedColumns.some(isUnresolvableColumnReference)) {
        return;
    }
    try {
        const resultTypeAssertion = invocation.resultTypeAssertion;
        const resultSchema = resolveColumnRefsToTableSchemaDescriptor(invocation.query.returnedColumns, tables);
        validateQueryMatchesReturnType(resultSchema, resultTypeAssertion.schema, resultTypeAssertion.path, invocation.query);
    }
    catch (error) {
        throw errors_1.augmentFileValidationError(error, invocation.query);
    }
}
function validateQuery(invocation, tables) {
    const tableExpressions = utils_1.getAllSubqueries(invocation.query)
        .map(subquery => subquery.exposedAsTable
        ? resolveSubqueryToTableSchema(subquery, tables)
        : null)
        .filter((schemaOrNull) => schemaOrNull !== null);
    validateSubquery(invocation.query, [...tables, ...tableExpressions]);
    validateQueryInvocation(invocation, [...tables, ...tableExpressions]);
}
exports.validateQuery = validateQuery;
